# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements total/partial Binary to Thermometer decoder."""

from tensorflow.keras.utils import to_categorical
import numpy as np


def BinaryToThermometer(
    x, classes, value_range, with_residue=False, merge_with_channels=False,
    use_two_hot_encoding=False):

  """Converts binary to one-hot (with scales).

  Given input matrix x with values (for example) 0, 1, 2, 3, 4, 5, 6, 7, create
  a number of classes as follows:

  classes=2, value_range=8, with_residue=0

  A true one-hot representation, and the remaining bits are truncated, using
  one bit representation.

  0 - [1,0] 1 - [1,0] 2 - [1,0] 3 - [1,0]
  4 - [0,1] 5 - [0,1] 6 - [0,1] 7 - [0,1]

  classes=2, value_range=8, with_residue=1

  In this case, the residue is added to the one-hot class, and the class will
  use 2 bits (for the remainder) + 1 bit (for the one hot)

  0 - [1,0] 1 - [1.25,0] 2 - [1.5,0] 3 - [1.75,0]
  4 - [0,1] 5 - [0,1.25] 6 - [0,1.5] 7 - [0,1.75]

  Arguments:
    x: the input vector we want to convert. typically its dimension will be
      (B,H,W,C) for an image, or (B,T,C) or (B,C) for for a 1D signal, where
      B=batch, H=height, W=width, C=channels or features, T=time for time
      series.
    classes: the number of classes to (or log2(classes) bits) to use of the
      values.
    value_range: max(x) - min(x) over all possible x values (e.g. for 8 bits,
      we would use 256 here).
    with_residue: if true, we split the value range into two sets and add
      the decimal fraction of the set to the one-hot representation for partial
      thermometer representation.
    merge_with_channels: if True, we will not create a separate dimension
      for the resulting matrix, but we will merge this dimension with
      the last dimension.
    use_two_hot_encoding: if true, we will distribute the weight between
      the current value and the next one to make sure the numbers will always
      be < 1.

  Returns:
    Converted x with classes with the last shape being C*classes.

  """

  # just make sure we are processing floats so that we can compute fractional
  # values

  x = x.astype(np.float32)

  # the number of ranges are equal to the span of the original values
  # divided by the number of target classes.
  #
  # for example, if value_range is 256 and number of classes is 16, we have
  # 16 values (remaining 4 bits to redistribute).

  ranges = value_range/classes
  x_floor = np.floor(x / ranges)

  if use_two_hot_encoding:
    x_ceil = np.ceil(x / ranges)

  if with_residue:
    x_mod_f = (x - x_floor * ranges) / ranges

  # convert values to categorical. if use_two_hot_encoding, we may
  # end up with one more class because we need to distribute the
  # remaining bits to the saturation class. For example, if we have
  # value_range = 4 (0,1,2,3) and classes = 2, if we use_two_hot_encoding
  # we will have the classes 0, 1, 2, where for the number 3, we will
  # allocate 0.5 to bin 1 and 0.5 to bin 2 (namelly 3 = 0.5 * (2**2 + 2**1)).

  xc_f = to_categorical(x_floor, classes + use_two_hot_encoding)

  if with_residue:
    xc_f_m = xc_f == 1

    if use_two_hot_encoding:
      xc_c = to_categorical(x_ceil, classes + use_two_hot_encoding)
      xc_c_m = xc_c == 1
      if np.any(xc_c_m):
        xc_c[xc_c_m] = x_mod_f.reshape(xc_c[xc_c_m].shape)
      if np.any(xc_f_m):
        xc_f[xc_f_m] = (1.0 - x_mod_f.reshape(xc_f[xc_f_m].shape))
      xc_f += xc_c
    else:
      if np.any(xc_f_m):
        xc_f[xc_f_m] += x_mod_f.reshape(xc_f[xc_f_m].shape)

  if merge_with_channels and len(xc_f.shape) != len(x.shape):
    sz = xc_f.shape
    sz = sz[:-2] + (sz[-2] * sz[-1],)
    xc_f = xc_f.reshape(sz)

  return xc_f




# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Divider operation implementation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import numpy as np


class IDividerImpl(abc.ABC):
  """abstract class for divider."""

  def __init__(self, numerator_quantizer, denominator_quantizer,
               output_quantizer):
    self.numerator_quantizier = numerator_quantizer
    self.denominator_quantizer = denominator_quantizer
    self.output = output_quantizer

  @staticmethod
  @abc.abstractmethod
  def implemented_as():
    pass


class FloatingPointDivider(IDividerImpl):
  """floating point divider."""

  def __init__(self, numerator_quantizer, denominator_quantizer,
               output_quantizer):

    super().__init__(numerator_quantizer, denominator_quantizer,
                     output_quantizer)
    if self.output.bits is None:
      # decide f16/f32 according to numerator/denominator type
      bits = 0
      if numerator_quantizer.is_floating_point:
        bits = max(bits, numerator_quantizer.bits)
      if denominator_quantizer.is_floating_point:
        bits = max(bits, denominator_quantizer.bits)

      self.output.bits = bits

    self.gate_bits = self.output.bits
    self.gate_factor = 1

  @staticmethod
  def implemented_as():
    # TODO(lishanok): change cost from "mul" to "divide"
    return "mul"


class Shifter(IDividerImpl):
  """shifter type."""

  # other_datatype/po2
  def __init__(self, numerator_quantizer, denominator_quantizer,
               output_quantizer):
    super().__init__(numerator_quantizer, denominator_quantizer,
                     output_quantizer)

    qbit_quantizer = numerator_quantizer
    po2_quantizer = denominator_quantizer

    (min_exp, max_exp) = po2_quantizer.get_min_max_exp()

    # since it's a divider, min_exp and max_exp swap
    # for calculating right and left shift
    tmp = min_exp
    min_exp = max_exp
    max_exp = tmp

    qbits_bits = qbit_quantizer.bits
    qbits_int_bits = qbit_quantizer.int_bits

    self.output.bits = int(qbits_bits + max_exp + min_exp)
    if (not qbit_quantizer.is_signed) and po2_quantizer.is_signed:
      # if qbit is signed, qbits_bits already has the sign_bit,
      # no need to +1,
      # if qbit is un_signed, po2 is unsigned, no need to +1
      # if qbit is un_signed, po2 is signed, min_exp and max_exp
      # didnot include sign_bit,
      # therefore need to +1
      self.output.bits += 1

    self.output.int_bits = int(qbits_int_bits + max_exp)
    self.output.is_signed = qbit_quantizer.is_signed |\
                            po2_quantizer.is_signed
    self.output.is_floating_point = False

    if po2_quantizer.inference_value_counts > 0:
      # during qbn inference, count number of unique values
      self.gate_factor = po2_quantizer.inference_value_counts * 0.3
      self.gate_bits = qbits_bits
    else:
      # programmable shifter, similar to sum gate
      self.gate_factor = 1
      b = np.sqrt(2 ** po2_quantizer.bits * qbits_bits)
      self.gate_bits = b * np.log10(b)

  @staticmethod
  def implemented_as():
    return "shifter"


class Subtractor(IDividerImpl):
  """subtractor quantizer."""

  # subtractor is only possible when numerator and denominator
  # are both po2 quantizers.

  def __init__(self, numerator_quantizer, denominator_quantizer,
               output_quantizer):
    super().__init__(numerator_quantizer, denominator_quantizer,
                     output_quantizer)

    self.output.bits = max(numerator_quantizer.bits,
                           denominator_quantizer.bits) + 1
    self.output.int_bits = max(numerator_quantizer.int_bits,
                               denominator_quantizer.int_bits) + 1
    self.output.is_signed = 1
    self.output.is_floating_point = False
    self.output.is_po2 = 1

    if (numerator_quantizer.max_val_po2 == -1 or
        denominator_quantizer.max_val_po2 == -1):
      self.output.max_val_po2 = -1
    else:
      # Adder is two po2_value multiply with each other
      self.output.max_val_po2 = numerator_quantizer.max_val_po2 /\
                                denominator_quantizer.max_val_po2

    if "po2" in output_quantizer.name:
      # po2 * po2
      if self.output.is_signed:
        output_quantizer.name = "quantized_po2"
      else:
        output_quantizer.name = "quantized_relu_po2"

    self.gate_bits = self.output.bits
    self.gate_factor = 1

  @staticmethod
  def implemented_as():
    return "add"



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""decides which quantizer implementation to use."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
from qkeras import quantizers


from qkeras.qtools.quantized_operators import quantizer_impl
from qkeras.qtools.settings import cfg


class QuantizerFactory:
  """Convert qkeras quantizer to qtools quantizer type."""

  def __init__(self):
    self.quantizer_lookup = {
        quantizers.quantized_bits:
            quantizer_impl.QuantizedBits,
        quantizers.binary:
            quantizer_impl.Binary,
        quantizers.quantized_relu:
            quantizer_impl.QuantizedRelu,
        quantizers.ternary:
            quantizer_impl.Ternary,
        quantizers.quantized_relu_po2:
            quantizer_impl.ReluPowerOfTwo,
        quantizers.quantized_po2:
            quantizer_impl.PowerOfTwo,
        quantizers.stochastic_ternary:
            quantizer_impl.StochasticTernary,
        quantizers.stochastic_binary:
            quantizer_impl.StochasticBinary,
        quantizers.bernoulli:
            quantizer_impl.Bernoulli,
        quantizers.quantized_tanh:
            quantizer_impl.QuantizedTanh,
        quantizers.quantized_ulaw:
            quantizer_impl.QuantizedUlaw,
        
            

        # add following quantizer types for the use in GraphUpdateEdge
        quantizer_impl.QuantizedBits:
            quantizer_impl.QuantizedBits,
        quantizer_impl.Binary:
            quantizer_impl.Binary,
        quantizer_impl.QuantizedRelu:
            quantizer_impl.QuantizedRelu,
        quantizer_impl.Ternary:
            quantizer_impl.Ternary,
        quantizer_impl.ReluPowerOfTwo:
            quantizer_impl.ReluPowerOfTwo,
        quantizer_impl.PowerOfTwo:
            quantizer_impl.PowerOfTwo,
        quantizer_impl.FloatingPoint:
            quantizer_impl.FloatingPoint,
        quantizer_impl.StochasticTernary:
            quantizer_impl.StochasticTernary,
        quantizer_impl.StochasticBinary:
            quantizer_impl.StochasticTernary,
        quantizer_impl.Bernoulli:
            quantizer_impl.StochasticTernary,
        quantizer_impl.QuantizedTanh:
            quantizer_impl.StochasticTernary,
        quantizer_impl.QuantizedUlaw:
            quantizer_impl.StochasticTernary,
        
            
    }

    self._default_interm_quantizer = cfg.default_interm_quantizer

  def _make_quantizer_util(self, quantizer) -> quantizer_impl.IQuantizer:
    """make quantizer util function."""
    if quantizer in ["int8", "int16", "int32", "fp16", "fp32"]:
      return self.make_default_quantizer(mode=quantizer)

    elif isinstance(quantizer, tuple(self.quantizer_lookup.keys())):
      quantizer_class = self.quantizer_lookup[type(quantizer)]
      if quantizer_class == type(quantizer):
        return self.clone_quantizer(quantizer)
      else:
        q = quantizer_class()
        q.convert_qkeras_quantizer(quantizer)
        return q

    return None

  def make_quantizer(self, quantizer) -> quantizer_impl.IQuantizer:
    """create quantizer according to input qkeras quantizer."""

    q = None
    if quantizer is not None:
      q = self._make_quantizer_util(quantizer)

    if q is None:
      return self.make_default_quantizer(
          mode=self._default_interm_quantizer)

    return q

  def is_quantizer_supported(self, quantizer) -> bool:
    if quantizer is None:
      # if None, will use default quantizer defined in config.json
      return True

    return isinstance(quantizer, tuple(self.quantizer_lookup.keys()))

  def make_default_quantizer(self, mode) -> quantizer_impl.IQuantizer:
    """make quantizer given qkeras quantizer type."""
    if mode == "fp32":
      return quantizer_impl.FloatingPoint(
          bits=32)
    elif mode == "fp16":
      return quantizer_impl.FloatingPoint(
          bits=16)
    elif mode == "int8":
      qbits = quantizer_impl.QuantizedBits()
      qbits.convert_qkeras_quantizer(
          quantizers.quantized_bits(8, 0, 1))
      return qbits
    elif mode == "int16":
      qbits = quantizer_impl.QuantizedBits()
      qbits.convert_qkeras_quantizer(
          quantizers.quantized_bits(16, 7, 1))
      return qbits
    elif mode == "int32":
      qbits = quantizer_impl.QuantizedBits()
      qbits.convert_qkeras_quantizer(
          quantizers.quantized_bits(32, 10, 1))
      return qbits
    else:
      try:
        # string to quantizer object
        q_name = "quantizers." + mode
        qkeras_object = eval(q_name)  # pylint: disable=eval-used
        return self._make_quantizer_util(qkeras_object)
      except:  # pylint: disable=bare-except
        raise ValueError("unaccepted quantizer {}!".format(mode))

  def clone_quantizer(
      self, quantizer: quantizer_impl.IQuantizer) -> quantizer_impl.IQuantizer:
    """clone the given quantizer."""
    return copy.deepcopy(quantizer)



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Creates networkx graph from a model."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging

import networkx as nx
import tensorflow.keras.backend as K

from qkeras.qtools.quantized_operators import quantizer_factory as quantizer_factory_module
from qkeras.qtools.settings import cfg
from tensorflow.keras.layers import InputLayer

SOURCE = -1
SINK = -2


class WrongInputQuantizerError(ValueError):
  pass


def GraphRemoveNode(graph, v):
  """Removes node "v" from u -> v -> w, connecting u -> w."""

  incoming = [u for u in graph.predecessors(v) if u != v]
  outgoing = [w for w in graph.successors(v) if w != v]

  # add incoming edges
  for u in incoming:
    for w in outgoing:
      in_attr = graph[u][v]
      out_attr = graph[v][w]

      assert list(in_attr["shape"]) == list(out_attr["shape"])

      graph.add_edges_from([(u, w, out_attr)])

  graph.remove_node(v)


def GraphRemoveNodeWithNodeType(graph, node_type):
  """Removes node with attribute node_type, reconnecting network."""

  nodes_to_remove = [v for v in graph.nodes
                     if graph.nodes[v]["type"][-1] == node_type]

  for v in nodes_to_remove:

    GraphRemoveNode(graph, v)


def  GraphAddHiddenInputLayer(model, graph, input_quantizer_map):
  """For Keras Sequential model api, input layer is hidden. Need to add it."""

  node_id = -1
  for (u, _) in graph.nodes.items():
    if u >= node_id:
      node_id = u
    if u == SOURCE or u == SINK:
      continue

    if graph.nodes[u]["type"][-1] == "InputLayer":
      return

  # determine a node id for the newly added input layer
  node_id += 1

  # find the first layer of the sequential model
  first_layer_nodes = []
  for u in graph.nodes:
    if u == SOURCE or u == SINK:
      continue
    predecessors = list(graph.predecessors(u))
     # find the first layer which doesn't have a parent
    if not predecessors:
      first_layer_nodes.append(u)
  assert len(first_layer_nodes) == 1
  # since it is a sequential model, there is only one first layer
  v_id = first_layer_nodes[0]

  # create a input layer node
  node_type = "InputLayer"
  input_shape = model.layers[0].input_shape
  layer = InputLayer(input_shape=input_shape[1:])
  o_shape = input_shape
  node = (node_id, {"layer": [layer], "type": [node_type],
                    "out_quantizer": None})
  graph.add_nodes_from([node])

  # insert input_quantizers on the edge between input layer and its next layer
  for (a, _) in input_quantizer_map.items():
    edge = (node_id, v_id, {
        "shape": [o_shape], "tensor": a,
        "quantizer": input_quantizer_map[a]})

  graph.add_edges_from([edge])


def GraphAddSingleSourceSingleSink(graph):

  """Connects graph to source and sink nodes."""

  edge_list = []

  for u in graph.nodes:

    if u == SOURCE or u == SINK:
      continue

    if graph.nodes[u]["type"][-1] == "InputLayer":
      # If the layer has multiple nodes, you can use get_output_at(node_index)
      tensor = graph.nodes[u]["layer"][-1].get_output_at(0)
      # if tf 1.0+, we can do tensor.shape with the same effect
      shape = tuple(tensor.get_shape().as_list())
      shape = [shape]

      edge_list.append((SOURCE, u, {
          "shape": shape, "tensor": tensor, "quantizer": None}))

    if graph.out_degree(u) == 0:
      tensor = graph.nodes[u]["layer"][-1].get_output_at(0)
      shape = tensor.shape

      edge_list.append((u, SINK, {
          "shape": shape, "tensor": tensor, "quantizer": None}))

  graph.add_edges_from(edge_list)


def GenerateGraphFromModel(model, input_quantizers,
                           default_source_quantizer):
  """Generates single source, single sink graph from model."""

  # node represents layers with attributes [layer, type(class_name)]
  # edge represents the tensor flowing between two layers,
  # attributes is [tensor, output_shape, QA(activation quantizer]

  # input_quantizers are tagged on the edge between input
  # layer and the following layer

  # generate a list of input quantizers
  input_quantizer_list = []
  quantizer_factory = quantizer_factory_module.QuantizerFactory()
  if input_quantizers is None:
    logging.warning(
        "************ SOURCE has no quantizer type."
        " Use default quantizer instead")

    for _ in range(len(model.inputs)):
      input_quantizer_list.append(
          quantizer_factory.make_default_quantizer(
              mode=default_source_quantizer))
  else:
    if len(model.inputs) == len(input_quantizers):
      for quantizer in input_quantizers:
        input_quantizer_list.append(quantizer_factory.make_quantizer(
            quantizer))
    # pass a single quantizer which will be used for all q list.
    elif not isinstance(input_quantizers, list):
      for _ in range(len(model.inputs)):
        input_quantizer_list.append(quantizer_factory.make_quantizer(
            input_quantizers))
    else:
      raise WrongInputQuantizerError(
          "ERROR: Numer of input (%d) must be the same as number of source"
          " quantizers (%d)"%(len(model.inputs), len(input_quantizers)))

  # dict that map input_tensor to its quantizer
  input_quantizer_map = {}
  for (idx, tensor) in enumerate(model.inputs):
    input_quantizer_map[tensor.experimental_ref()] = input_quantizer_list[idx]

  graph = nx.DiGraph()

  source = SOURCE
  sink = SINK

  node_list = [
      (source, {"layer": [None], "type": [None], "out_quantizer": None}),
      (sink, {"layer": [None], "type": [None], "out_quantizer": None})
  ]

  graph.add_nodes_from(node_list)

  node_list = []

  for i, layer in enumerate(model.layers):

    node_type = layer.__class__.__name__

    node = (i, {"layer": [layer], "type": [node_type], "out_quantizer": None})
    node_list.append(node)

  node_dict = {layer: i for i, layer in enumerate(model.layers)}

  graph.add_nodes_from(node_list)

  # nodes = tensors
  in_nodes = {}
  out_nodes = {}
  for layer in model.layers:
    i_list = layer.input
    if not isinstance(layer.input, list):
      i_list = [i_list.experimental_ref()]
    else:
      i_list = [tmp.experimental_ref() for tmp in i_list]

    for i in i_list:
      # dict: tensor -> layers have this tensor as input
      if i not in in_nodes.keys():
        in_nodes[i] = [layer]
      else:
        in_nodes[i].append(layer)

    o_list = layer.output
    if not isinstance(layer.output, list):
      o_list = [o_list.experimental_ref()]
    else:
      o_list = [tmp.experimental_ref() for tmp in o_list]

    for o in o_list:
      # dict: tensor -> layer have this tensor as output
      if o not in out_nodes.keys():
        out_nodes[o] = [layer]
      else:
        out_nodes[o].append(layer)

  # union of all tensors; non-redundant
  attr_set = set(in_nodes.keys()) | set(out_nodes.keys())

  # add edges. we want edges annotated with tensors and shapes

  edge_list = []

  for a in attr_set:
    # for a given tensor a, find the layer u that outputs this tensor
    # and the layer v that has this tensor as input
    u_list = out_nodes.get(a, [None])
    v_list = in_nodes.get(a, [None])

    for u in u_list:
      for v in v_list:
        if not u or not v:
          continue

        o_shape = u.output_shape

        # layer -> layer_id
        u_id = node_dict[u]
        v_id = node_dict[v]

        # insert input_quantizers on the edge between
        # input layer and its next layer
        if a in input_quantizer_map.keys():
          edge_list.append((u_id, v_id, {
              "shape": o_shape, "tensor": a,
              "quantizer": input_quantizer_map[a]}))
        else:
          edge_list.append((u_id, v_id, {
              "shape": o_shape, "tensor": a,
              "quantizer": None}))

  graph.add_edges_from(edge_list)
  GraphAddHiddenInputLayer(model, graph, input_quantizer_map)

  return (graph, input_quantizer_list)


def GraphGetInputs(graph):

  """Returns edges SOURCE->u that are inputs."""

  successors = list(graph.successors(SOURCE))

  input_tensors = []

  for u in successors:

    if u == SOURCE or u == SINK:
      continue

    input_tensors.append(graph[SOURCE][u])

  return input_tensors


def GraphGetOutputs(graph):

  """Returns edges u->SINK that are outputs."""

  predecessors = list(graph.predecessors(SINK))

  output_tensors = []

  for u in predecessors:

    if u == SOURCE or u == SINK:
      continue

    output_tensors.append(graph[u][SINK])

  return output_tensors


def GraphPropagateActivationsToEdges(graph, debug=False):
  """Traverses graph and move activations to edges.

  1.If current dense/conv layer is specified with QA:
    outgoing edge (output data type) will be QA type
  2.If current dense/conv layer has no QA:
    default type (float32) is used as output
  3.If current layer is QA layer:
    float32 is used by default as output type on the edge

  Args:
    graph: graph to inject activations to.
    debug: debug mode

  Returns:
    None
  """

  scheduler = list(nx.topological_sort(graph))

  for vertex in scheduler[1:-1]:
    # get rid of source and sink vertex
    if debug:
      print("########### GraphPropagateActivationsToEdges ############")
      print("vertex:", vertex)

    for u, v in graph.edges(vertex):
      # u=vertex, v: outgoing edge vertex

      if debug:
        print("  outgoing ->", v, graph.nodes[v]["layer"][0].name)

      layer = graph.nodes[u]["layer"][0]
      result = None
      # if current layer has no QA specified
      if not hasattr(layer, "activation"):
        result = None
      else:
        activation_name = layer.activation.__name__ if hasattr(
            layer.activation, "__name__") else None
        q_activation_class_name = layer.activation.__class__.__name__ if hasattr(
            layer.activation, "__class__") else None

        if debug:
          print("  layer type:", layer.__class__.__name__)
          print("  activation object:", layer.activation)
          print("  activation_name:", activation_name)
          print("  q_activation_class_name:", q_activation_class_name)

        # if current layer is QA
        if (graph.nodes[u]["type"][0] in ["QActivation"] or
            graph.nodes[u]["type"][0] in ["QAdaptiveActivation"]):
          result = layer.quantizer

        # if current layer is not QA layer but has QA specified within
        elif hasattr(layer, "activation"):
          if activation_name == "linear":
            result = None
          else:
            result = layer.activation

      if debug:
        print("  {}->{}: {}".format(u, v, result))

      graph[u][v]["quantizer"] = result
      # all edge_quantizer is the same for all edges starting
      # from current vertex to different nodes
      graph.nodes[vertex]["out_quantizer"] = result


def PrintGraph(graph, msg=""):
  """Print graph structure."""

  print()
  print(msg)
  print()
  print("nodes:",
        [(u, graph.nodes[u]["layer"][
            0].name if graph.nodes[u]["layer"][0] is not None else "",
          graph.nodes[u]["type"]) for u in graph.nodes])
  print()
  print("edges:",
        [(u, v, graph[u][v]["shape"],
          graph[u][v]["quantizer"]) for u, v in graph.edges])


def CreateGraph(model, input_quantizers=None,
                default_source_quantizer=cfg.default_source_quantizer,
                debug=False):
  """create graph."""

  K.set_image_data_format("channels_last")

  (graph, source_quantizer_list) = GenerateGraphFromModel(
      model, input_quantizers, default_source_quantizer)
  GraphAddSingleSourceSingleSink(graph)
  GraphRemoveNodeWithNodeType(graph, "Dropout")
  GraphRemoveNodeWithNodeType(graph, "InputLayer")

  scheduler = list(nx.topological_sort(graph))

  if debug:
    for vertex in scheduler[1:-1]:
      for _, v in graph.edges(vertex):
        if v == SINK:
          continue
        print("... calling", graph.nodes[v][
            "layer"][0].name, graph.nodes[v]["type"])

  return (graph, source_quantizer_list)


def GraphUpdateEdge(graph, node_id, quantizer_on_edge):
  """update the graph edges outgoing from node_id with new quantizer."""

  for u, v in graph.edges(node_id):
    graph[u][v]["quantizer"] = quantizer_on_edge



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""configuration file for external usage."""

config_settings = {
    "default_source_quantizer": "quantized_bits(8, 0, 1)",
    "default_interm_quantizer": "quantized_bits(8, 0, 1)",

    "horowitz": {
        "fpm_add": [0.003125, 0],
        "fpm_mul": [0.002994791667, 0.001041666667, 0],
        "fp16_add": [0.4],
        "fp16_mul": [1.1],
        "fp32_add": [0.9],
        "fp32_mul": [3.7],
        "sram_rd": [9.02427321e-04, -2.68847858e-02, 2.08900804e-01, 0.0],
        "dram_rd": [20.3125, 0]
    },

    "include_energy": {
        "QActivation": ["outputs"],
        "QAdaptiveActivation": ["outputs"],
        "Activation": ["outputs"],
        "QBatchNormalization": ["parameters"],
        "BatchNormalization": ["parameters"],
        "Add": ["op_cost"],
        "Subtract": ["op_cost"],
        "MaxPooling2D": ["op_cost"],
        "default": ["inputs", "parameters", "op_cost"]
    }
}



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Interface for running qtools and qenergy."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json

from qkeras.qtools import generate_layer_data_type_map
from qkeras.qtools import interface
from qkeras.qtools import qgraph
from qkeras.qtools.config_public import config_settings
from qkeras.qtools.qenergy import qenergy
from qkeras.qtools.settings import cfg


class QTools:
  """integration of different qtools functions."""

  def __init__(self, model, process, source_quantizers=None,
               is_inference=False, weights_path=None,
               keras_quantizer=None, keras_accumulator=None,
               for_reference=False):

    if model is not None:
      self._model = model

    if weights_path is not None:
      self._model.load_weights(weights_path)

    cfg.update(process, config_settings)

    # if source_quantizers is None, CreateGraph will use
    # default_source_quantizers defined in cfg
    (graph, source_quantizer_list) = qgraph.CreateGraph(
        model, source_quantizers, cfg.default_source_quantizer)

    # qgraph.PrintGraph(graph)
    qgraph.GraphPropagateActivationsToEdges(graph)
    self._layer_map = generate_layer_data_type_map.generate_layer_data_type_map(
        graph, source_quantizer_list, is_inference,
        keras_quantizer, keras_accumulator, for_reference)

    self._output_dict = interface.map_to_json(self._layer_map)

  def qtools_stats_to_json(self, json_name):
    """dump the layer stats to a json file."""

    with open(json_name, "w") as outfile:
      json.dump(self._output_dict, outfile, indent=4)

  def qtools_stats_print(self):
    """print out the layer stats."""

    dict_to_json = json.dumps(self._output_dict, indent=4)
    print(dict_to_json)

  def pe(self, weights_on_memory="dram",
         activations_on_memory="dram",
         min_sram_size=0,
         rd_wr_on_io=True,
         verbose=False):
    """energy consumption calculation."""

    assert weights_on_memory in ["dram", "sram", "fixed"]
    energy_dict = qenergy.energy_estimate(
        self._model, self._layer_map, weights_on_memory,
        activations_on_memory, min_sram_size,
        rd_wr_on_io)

    if verbose:
      print("COST:")
      dict_to_json = json.dumps(energy_dict, indent=4)
      print(dict_to_json)

    return energy_dict

  def extract_energy_sum(self, cfg_setting, energy_dict):
    """extracted energy needed in caculating sum."""

    value = 0
    for layer in energy_dict.keys():
      if layer == "total_cost":
        continue

      class_name = energy_dict[layer]["class_name"]
      keys = cfg_setting.get(class_name, cfg_setting.get("default", []))
      value += sum([energy_dict[layer]["energy"][key] for key in keys])

    return int(value)

  def extract_energy_profile(self, cfg_setting, energy_dict):
    """extract energy consumption in each layer."""

    energy = {}
    for layer in energy_dict.keys():
      if layer == "total_cost":
        continue

      class_name = energy_dict[layer]["class_name"]
      keys = cfg_setting.get(class_name, cfg_setting.get("default", []))
      energy[layer] = {}
      energy[layer]["energy"] = energy_dict[layer]["energy"]
      energy[layer]["total"] = sum(
          [energy_dict[layer]["energy"][key] for key in keys])

    return energy



# Lint as: python3
# ==============================================================================
# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements utility functions for support of auto-quantization."""

import json
import tensorflow as tf


Q_SEQUENCE_LAYERS = ["QSimpleRNN", "QLSTM", "QGRU", "QBidirectional"]

def print_qmodel_summary(q_model):
  """Prints quantized model summary."""

  for layer in q_model.layers:
    if (layer.__class__.__name__ == "QActivation" or
        layer.__class__.__name__ == "QAdaptiveActivation"):
      print("{:20} {}".format(layer.name, str(layer.activation)))
    elif (
        hasattr(layer, "get_quantizers") and
        layer.__class__.__name__ != "QBatchNormalization"
    ):
      print("{:20} ".format(layer.name), end="")
      if "Dense" in layer.__class__.__name__:
        print("u={} ".format(layer.units), end="")
      elif layer.__class__.__name__ in [
          "Conv2D", "QConv2D", "Conv1D", "QConv1D",
          "QConv2DBatchnorm", "QDepthwiseConv2DBatchnorm"]:
        print("f={} ".format(layer.filters), end="")
      quantizers = layer.get_quantizers()
      for q in range(len(quantizers)):
        if quantizers[q] is not None:
          print("{} ".format(str(quantizers[q])), end="")
      if hasattr(layer, "recurrent_activation"):
        print("recurrent act={}".format(layer.recurrent_activation), end="")
      if (
          layer.activation is not None and
          not (
              hasattr(layer.activation, "__name__") and
              layer.activation.__name__ == "linear"
          )
      ):
        print("act={}".format(layer.activation), end="")
      print()
    elif layer.__class__.__name__ == "QBatchNormalization":
      print("{:20} QBN, mean={}".format(layer.name,
          str(tf.keras.backend.eval(layer.moving_mean))), end="")
      print()
    elif layer.__class__.__name__ == "BatchNormalization":
      print("{:20} is normal keras bn layer".format(layer.name), end="")
      print()

  print()


def get_quantization_dictionary(q_model):
  """Returns quantization dictionary."""

  q_dict = {}
  for layer in q_model.layers:
    if hasattr(layer, "get_quantization_config"):
      q_dict[layer.name] = layer.get_quantization_config()

  return q_dict


def save_quantization_dict(fn, q_model):
  """Saves quantization dictionary as json object in disk."""
  q_dict = get_quantization_dictionary(q_model)
  json_dict = json.dumps(q_dict)

  f = open(fn, "w")
  f.write(json_dict + "\n")
  f.close()

